@@ -19,8 +19,11 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
+#include "libavutil/channel_layout.h"
 #include "avcodec.h"
 #include "bytestream.h"
+#include "codec_internal.h"
+#include "encode.h"
 #include "internal.h"
 
 typedef struct PCMDVDContext {
@@ -28,8 +31,6 @@ typedef struct PCMDVDContext {
     int block_size;          // Size of a block of samples in bytes
     int samples_per_block;   // Number of samples per channel per block
     int groups_per_block;    // Number of 20/24-bit sample groups per block
-    uint8_t *extra_samples;  // Pointer to leftover samples from a frame
-    int extra_sample_count;  // Number of leftover samples in the buffer
 } PCMDVDContext;
 
 static av_cold int pcm_dvd_encode_init(AVCodecContext *avctx)
@@ -44,6 +45,8 @@ static av_cold int pcm_dvd_encode_init(AVCodecContext *avctx)
     case 96000:
         freq = 1;
         break;
+    default:
+        av_assert1(0);
     }
 
     switch (avctx->sample_fmt) {
@@ -55,10 +58,12 @@ static av_cold int pcm_dvd_encode_init(AVCodecContext *avctx)
         avctx->bits_per_coded_sample = 24;
         quant = 2;
         break;
+    default:
+        av_assert1(0);
     }
 
     avctx->bits_per_coded_sample = 16 + quant * 4;
-    avctx->block_align           = avctx->channels * avctx->bits_per_coded_sample / 8;
+    avctx->block_align           = avctx->ch_layout.nb_channels * avctx->bits_per_coded_sample / 8;
     avctx->bit_rate              = avctx->block_align * 8LL * avctx->sample_rate;
     if (avctx->bit_rate > 9800000) {
         av_log(avctx, AV_LOG_ERROR, "Too big bitrate: reduce sample rate, bitdepth or channels.\n");
@@ -67,16 +72,16 @@ static av_cold int pcm_dvd_encode_init(AVCodecContext *avctx)
 
     if (avctx->sample_fmt == AV_SAMPLE_FMT_S16) {
         s->samples_per_block = 1;
-        s->block_size        = avctx->channels * 2;
+        s->block_size        = avctx->ch_layout.nb_channels * 2;
         frame_size           = 2008 / s->block_size;
     } else {
-        switch (avctx->channels) {
+        switch (avctx->ch_layout.nb_channels) {
         case 1:
         case 2:
         case 4:
             /* one group has all the samples needed */
             s->block_size        = 4 * avctx->bits_per_coded_sample / 8;
-            s->samples_per_block = 4 / avctx->channels;
+            s->samples_per_block = 4 / avctx->ch_layout.nb_channels;
             s->groups_per_block  = 1;
             break;
         case 8:
@@ -86,11 +91,11 @@ static av_cold int pcm_dvd_encode_init(AVCodecContext *avctx)
             s->groups_per_block  = 2;
             break;
         default:
-            /* need avctx->channels groups */
-            s->block_size        = 4 * avctx->channels *
+            /* need avctx->ch_layout.nb_channels groups */
+            s->block_size        = 4 * avctx->ch_layout.nb_channels *
                                    avctx->bits_per_coded_sample / 8;
             s->samples_per_block = 4;
-            s->groups_per_block  = avctx->channels;
+            s->groups_per_block  = avctx->ch_layout.nb_channels;
             break;
         }
 
@@ -98,7 +103,7 @@ static av_cold int pcm_dvd_encode_init(AVCodecContext *avctx)
     }
 
     s->header[0] = 0x0c;
-    s->header[1] = (quant << 6) | (freq << 4) | (avctx->channels - 1);
+    s->header[1] = (quant << 6) | (freq << 4) | (avctx->ch_layout.nb_channels - 1);
     s->header[2] = 0x80;
 
     if (!avctx->frame_size)
@@ -111,7 +116,7 @@ static int pcm_dvd_encode_frame(AVCodecContext *avctx, AVPacket *avpkt,
                                 const AVFrame *frame, int *got_packet_ptr)
 {
     PCMDVDContext *s = avctx->priv_data;
-    int samples = frame->nb_samples * avctx->channels;
+    int samples = frame->nb_samples * avctx->ch_layout.nb_channels;
     int64_t pkt_size = (frame->nb_samples / s->samples_per_block) * s->block_size + 3;
     int blocks = (pkt_size - 3) / s->block_size;
     const int16_t *src16;
@@ -119,7 +124,7 @@ static int pcm_dvd_encode_frame(AVCodecContext *avctx, AVPacket *avpkt,
     PutByteContext pb;
     int ret;
 
-    if ((ret = ff_alloc_packet2(avctx, avpkt, pkt_size, 0)) < 0)
+    if ((ret = ff_get_encode_buffer(avctx, avpkt, pkt_size, 0)) < 0)
         return ret;
 
     memcpy(avpkt->data, s->header, 3);
@@ -136,13 +141,13 @@ static int pcm_dvd_encode_frame(AVCodecContext *avctx, AVPacket *avpkt,
         } while (--samples);
         break;
     case AV_SAMPLE_FMT_S32:
-        if (avctx->channels == 1) {
+        if (avctx->ch_layout.nb_channels == 1) {
             do {
                 for (int i = 2; i; i--) {
                     bytestream2_put_be16(&pb, src32[0] >> 16);
                     bytestream2_put_be16(&pb, src32[1] >> 16);
-                    bytestream2_put_byte(&pb, (*src32++) >> 24);
-                    bytestream2_put_byte(&pb, (*src32++) >> 24);
+                    bytestream2_put_byte(&pb, (uint8_t)((*src32++) >> 8));
+                    bytestream2_put_byte(&pb, (uint8_t)((*src32++) >> 8));
                 }
             } while (--blocks);
         } else {
@@ -152,10 +157,10 @@ static int pcm_dvd_encode_frame(AVCodecContext *avctx, AVPacket *avpkt,
                     bytestream2_put_be16(&pb, src32[1] >> 16);
                     bytestream2_put_be16(&pb, src32[2] >> 16);
                     bytestream2_put_be16(&pb, src32[3] >> 16);
-                    bytestream2_put_byte(&pb, (*src32++) >> 24);
-                    bytestream2_put_byte(&pb, (*src32++) >> 24);
-                    bytestream2_put_byte(&pb, (*src32++) >> 24);
-                    bytestream2_put_byte(&pb, (*src32++) >> 24);
+                    bytestream2_put_byte(&pb, (uint8_t)((*src32++) >> 8));
+                    bytestream2_put_byte(&pb, (uint8_t)((*src32++) >> 8));
+                    bytestream2_put_byte(&pb, (uint8_t)((*src32++) >> 8));
+                    bytestream2_put_byte(&pb, (uint8_t)((*src32++) >> 8));
                 }
             } while (--blocks);
         }
@@ -163,35 +168,36 @@ static int pcm_dvd_encode_frame(AVCodecContext *avctx, AVPacket *avpkt,
     }
 
     avpkt->pts      = frame->pts;
-    avpkt->size     = pkt_size;
     avpkt->duration = ff_samples_to_time_base(avctx, frame->nb_samples);
     *got_packet_ptr = 1;
 
     return 0;
 }
 
-static av_cold int pcm_dvd_encode_close(AVCodecContext *avctx)
-{
-    return 0;
-}
-
-AVCodec ff_pcm_dvd_encoder = {
-    .name           = "pcm_dvd",
-    .long_name      = NULL_IF_CONFIG_SMALL("PCM signed 16|20|24-bit big-endian for DVD media"),
-    .type           = AVMEDIA_TYPE_AUDIO,
-    .id             = AV_CODEC_ID_PCM_DVD,
+const FFCodec ff_pcm_dvd_encoder = {
+    .p.name         = "pcm_dvd",
+    .p.long_name    = NULL_IF_CONFIG_SMALL("PCM signed 16|20|24-bit big-endian for DVD media"),
+    .p.type         = AVMEDIA_TYPE_AUDIO,
+    .p.id           = AV_CODEC_ID_PCM_DVD,
+    .p.capabilities = AV_CODEC_CAP_DR1 | AV_CODEC_CAP_SMALL_LAST_FRAME,
     .priv_data_size = sizeof(PCMDVDContext),
     .init           = pcm_dvd_encode_init,
-    .close          = pcm_dvd_encode_close,
-    .encode2        = pcm_dvd_encode_frame,
-    .capabilities   = AV_CODEC_CAP_SMALL_LAST_FRAME,
-    .supported_samplerates = (const int[]) { 48000, 96000, 0},
-    .channel_layouts = (const uint64_t[]) { AV_CH_LAYOUT_MONO,
+    FF_CODEC_ENCODE_CB(pcm_dvd_encode_frame),
+    .p.supported_samplerates = (const int[]) { 48000, 96000, 0},
+#if FF_API_OLD_CHANNEL_LAYOUT
+    .p.channel_layouts = (const uint64_t[]) { AV_CH_LAYOUT_MONO,
                                             AV_CH_LAYOUT_STEREO,
                                             AV_CH_LAYOUT_5POINT1,
                                             AV_CH_LAYOUT_7POINT1,
                                             0 },
-    .sample_fmts    = (const enum AVSampleFormat[]){ AV_SAMPLE_FMT_S16,
+#endif
+    .p.ch_layouts   = (const AVChannelLayout[]) { AV_CHANNEL_LAYOUT_MONO,
+                                                  AV_CHANNEL_LAYOUT_STEREO,
+                                                  AV_CHANNEL_LAYOUT_5POINT1,
+                                                  AV_CHANNEL_LAYOUT_7POINT1,
+                                                  { 0 } },
+    .p.sample_fmts  = (const enum AVSampleFormat[]){ AV_SAMPLE_FMT_S16,
                                                      AV_SAMPLE_FMT_S32,
                                                      AV_SAMPLE_FMT_NONE },
+    .caps_internal  = FF_CODEC_CAP_INIT_THREADSAFE,
 };
